import getpass
import json
import sys
import warnings

import click

from ai.backend.cli.main import main
from ai.backend.cli.types import ExitCode

from .. import __version__
from ..config import get_config, local_state_path
from ..exceptions import BackendClientError
from ..session import Session
from .pretty import print_done, print_error, print_fail, print_warn


@main.command()
def config():
    """
    Shows the current configuration.
    """
    config = get_config()
    click.echo(
        "API endpoint: {0} (mode: {1})".format(
            click.style(str(config.endpoint), bold=True),
            click.style(str(config.endpoint_type), fg="cyan", bold=True),
        )
    )
    click.echo(
        "Client version: {0} (API: {1})".format(
            click.style(__version__, bold=True),
            click.style(config.version, bold=True),
        )
    )
    if sys.stdout.isatty():
        click.echo("Server version: ...")
        click.echo("Negotiated API version: ...")
    else:
        with Session() as sess:
            try:
                versions = sess.System.get_versions()
            except BackendClientError:
                click.echo("Server version: (failed to fetch)")
            else:
                click.echo(
                    "Server version: {0} (API: {1})".format(
                        versions.get("manager", "pre-19.03"),
                        versions["version"],
                    )
                )
            click.echo("Negotiated API version: {0}".format(sess.api_version))
    nrows = 1
    if config.domain:
        click.echo('Domain name: "{0}"'.format(click.style(config.domain, bold=True)))
        nrows += 1
    if config.group:
        click.echo('Group name: "{0}"'.format(click.style(config.group, bold=True)))
        nrows += 1
    if config.is_anonymous:
        click.echo("Access key: (this is an anonymous session)")
        nrows += 1
    elif config.endpoint_type == "docker":
        pass
    elif config.endpoint_type == "session":
        if (local_state_path / "cookie.dat").exists() and (
            local_state_path / "config.json"
        ).exists():
            sess_config = json.loads((local_state_path / "config.json").read_text())
            click.echo(
                'Username: "{0}"'.format(click.style(sess_config.get("username", ""), bold=True))
            )
            nrows += 1
    else:
        click.echo('Access key: "{0}"'.format(click.style(config.access_key, bold=True)))
        nrows += 1
        masked_skey = config.secret_key[:6] + ("*" * 24) + config.secret_key[-10:]
        click.echo('Secret key: "{0}"'.format(click.style(masked_skey, bold=True)))
        nrows += 1
    click.echo("Signature hash type: {0}".format(click.style(config.hash_type, bold=True)))
    nrows += 1
    click.echo(
        "Skip SSL certificate validation? {0}".format(
            click.style(str(config.skip_sslcert_validation), bold=True)
        )
    )
    nrows += 1
    if sys.stdout.isatty():
        sys.stdout.flush()
        with warnings.catch_warnings(record=True) as captured_warnings, Session() as sess:
            click.echo("\u001b[{0}A\u001b[2K".format(nrows + 1), nl=False)
            try:
                versions = sess.System.get_versions()
            except BackendClientError:
                click.echo(
                    "Server version: {0}".format(
                        click.style("(failed to fetch)", fg="red", bold=True),
                    )
                )
            else:
                click.echo(
                    "Server version: {0} (API: {1})".format(
                        click.style(versions.get("manager", "pre-19.03"), bold=True),
                        click.style(versions["version"], bold=True),
                    )
                )
            click.echo("\u001b[2K", nl=False)
            click.echo(
                "Negotiated API version: {0}".format(
                    click.style("v{0[0]}.{0[1]}".format(sess.api_version), bold=True),
                )
            )
            click.echo("\u001b[{0}B".format(nrows), nl=False)
            sys.stdout.flush()
        for w in captured_warnings:
            warnings.showwarning(w.message, w.category, w.filename, w.lineno, w.line)


@main.command()
def login():
    """
    Log-in to the console API proxy.
    It stores the current session cookie in the OS-default
    local application data location.
    """
    user_id = input("User ID: ")
    password = getpass.getpass()

    config = get_config()
    if config.endpoint_type != "session":
        print_warn('To use login, your endpoint type must be "session".')
        raise click.Abort()

    with Session() as session:
        try:
            result = session.Auth.login(user_id, password)
            if (
                not result["authenticated"]
                and result.get("data", {}).get("details") == "OTP not provided"
            ):
                otp = input("One-time Password: ")
                result = session.Auth.login(user_id, password, otp=otp.strip())

            if not result["authenticated"]:
                print_fail("Login failed.")
                print_fail(result["data"]["details"])
                sys.exit(ExitCode.FAILURE)
            print_done("Login succeeded.")

            local_state_path.mkdir(parents=True, exist_ok=True)
            session.aiohttp_session.cookie_jar.update_cookies(result["cookies"])
            session.aiohttp_session.cookie_jar.save(local_state_path / "cookie.dat")
            (local_state_path / "config.json").write_text(json.dumps(result.get("config", {})))
        except Exception as e:
            print_error(e)


@main.command()
def logout():
    """
    Log-out from the console API proxy and clears the local cookie data.
    """
    config = get_config()
    if config.endpoint_type != "session":
        print_warn('To use logout, your endpoint type must be "session".')
        raise click.Abort()

    with Session() as session:
        try:
            session.Auth.logout()
            print_done("Logout done.")
            try:
                (local_state_path / "cookie.dat").unlink()
                (local_state_path / "config.json").unlink()
            except (IOError, PermissionError):
                pass
        except Exception as e:
            print_error(e)


@main.command()
@click.argument("old_password", metavar="OLD_PASSWORD")
@click.argument("new_password", metavar="NEW_PASSWORD")
@click.argument("new_password2", metavar="NEW_PASSWORD2")
def update_password(old_password, new_password, new_password2):
    """
    Update user's password.
    """
    config = get_config()
    if config.endpoint_type != "session":
        print_warn('To update password, your endpoint type must be "session".')
        raise click.Abort()

    with Session() as session:
        try:
            session.Auth.update_password(old_password, new_password, new_password2)
            print_done("Password updated.")
        except Exception as e:
            print_error(e)


@main.command()
@click.argument("domain", metavar="USER_ID")
@click.argument("user_id", metavar="USER_ID")
@click.argument("current_password", metavar="CURRENT_PASSWORD")
@click.argument("new_password", metavar="NEW_PASSWORD")
def update_password_no_auth(domain, user_id, current_password, new_password):
    """
    Update user's password. This is used to update `EXPIRED` password only.
    """
    with Session() as session:
        try:
            config = get_config()
            if config.endpoint_type == "session":
                session.Auth.update_password_no_auth_in_session(
                    user_id, current_password, new_password
                )
            else:
                session.Auth.update_password_no_auth(
                    domain, user_id, current_password, new_password
                )
            print_done("Password updated.")
        except Exception as e:
            print_error(e)



from typing import List, Mapping, Optional

from ..request import Request
from .base import BaseFunction, api_function

__all__ = ("Dotfile",)


class Dotfile(BaseFunction):
    @api_function
    @classmethod
    async def create(
        cls,
        data: str,
        path: str,
        permission: str,
        owner_access_key: Optional[str] = None,
        domain: Optional[str] = None,
        group: Optional[str] = None,
    ) -> "Dotfile":
        body = {
            "data": data,
            "path": path,
            "permission": permission,
        }
        if group:
            body["group"] = group
            if domain:
                body["domain"] = domain
            rqst_endpoint = "/group-config/dotfiles"
        elif domain:
            body["domain"] = domain
            rqst_endpoint = "/domain-config/dotfiles"
        else:
            if owner_access_key:
                body["owner_access_key"] = owner_access_key
            rqst_endpoint = "/user-config/dotfiles"

        rqst = Request("POST", rqst_endpoint)
        rqst.set_json(body)
        async with rqst.fetch() as resp:
            await resp.json()
        return cls(path, owner_access_key=owner_access_key, group=group, domain=domain)

    @api_function
    @classmethod
    async def list_dotfiles(
        cls,
        owner_access_key: Optional[str] = None,
        domain: Optional[str] = None,
        group: Optional[str] = None,
    ) -> "List[Mapping[str, str]]":
        params = {}
        if group:
            params["group"] = group
            if domain:
                params["domain"] = domain
            rqst_endpoint = "/group-config/dotfiles"
        elif domain:
            params["domain"] = domain
            rqst_endpoint = "/domain-config/dotfiles"
        else:
            if owner_access_key:
                params["onwer_access_key"] = owner_access_key
            rqst_endpoint = "/user-config/dotfiles"

        rqst = Request("GET", rqst_endpoint, params=params)
        async with rqst.fetch() as resp:
            return await resp.json()

    def __init__(
        self,
        path: str,
        owner_access_key: Optional[str] = None,
        group: Optional[str] = None,
        domain: Optional[str] = None,
    ):
        self.path = path
        self.owner_access_key = owner_access_key
        self.group = group
        self.domain = domain

    @api_function
    async def get(self) -> str:
        params = {"path": self.path}
        if self.group:
            params["group"] = self.group
            if self.domain:
                params["domain"] = self.domain
            rqst_endpoint = "/group-config/dotfiles"
        elif self.domain:
            params["domain"] = self.domain
            rqst_endpoint = "/domain-config/dotfiles"
        else:
            if self.owner_access_key:
                params["owner_access_key"] = self.owner_access_key
            rqst_endpoint = "/user-config/dotfiles"

        rqst = Request("GET", rqst_endpoint, params=params)
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    async def update(self, data: str, permission: str):
        body = {
            "data": data,
            "path": self.path,
            "permission": permission,
        }
        if self.group:
            body["group"] = self.group
            if self.domain:
                body["domain"] = self.domain
            rqst_endpoint = "/group-config/dotfiles"
        elif self.domain:
            body["domain"] = self.domain
            rqst_endpoint = "/domain-config/dotfiles"
        else:
            if self.owner_access_key:
                body["owner_access_key"] = self.owner_access_key
            rqst_endpoint = "/user-config/dotfiles"

        rqst = Request("PATCH", rqst_endpoint)
        rqst.set_json(body)
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    async def delete(self):
        params = {"path": self.path}
        if self.group:
            params["group"] = self.group
            if self.domain:
                params["domain"] = self.domain
            rqst_endpoint = "/group-config/dotfiles"
        elif self.domain:
            params["domain"] = self.domain
            rqst_endpoint = "/domain-config/dotfiles"
        else:
            if self.owner_access_key:
                params["owner_access_key"] = self.owner_access_key
            rqst_endpoint = "/user-config/dotfiles"

        rqst = Request("DELETE", rqst_endpoint, params=params)
        async with rqst.fetch() as resp:
            return await resp.json()



import json
import sys

import click

from ai.backend.cli.types import ExitCode

from ...session import Session
from ..pretty import print_error, print_fail, print_pretty
from . import admin


@admin.group()
def etcd() -> None:
    """
    etcd query and manipulation commands.
    (admin privilege required)
    """


@etcd.command()
@click.argument("key", type=str, metavar="KEY")
@click.option(
    "-p", "--prefix", is_flag=True, default=False, help="Get all keys prefixed with the given key."
)
def get(key, prefix):
    """
    Get a ETCD value(s).

    KEY: Name of ETCD key.
    """
    with Session() as session:
        try:
            data = session.EtcdConfig.get(key, prefix)
        except Exception as e:
            print_error(e)
            sys.exit(ExitCode.FAILURE)
        data = json.dumps(data, indent=2) if data else "null"
        print_pretty(data)


@etcd.command()
@click.argument("key", type=str, metavar="KEY")
@click.argument("value", type=str, metavar="VALUE")
def set(key, value):
    """
    Set new key and value on ETCD.

    KEY: Name of ETCD key.
    VALUE: Value to set.
    """
    with Session() as session:
        try:
            value = json.loads(value)
            print_pretty("Value converted to a dictionary.")
        except json.JSONDecodeError:
            pass
        try:
            data = session.EtcdConfig.set(key, value)
        except Exception as e:
            print_error(e)
            sys.exit(ExitCode.FAILURE)
        if data.get("result", False) != "ok":
            print_fail("Unable to set key/value.")
        else:
            print_pretty("Successfully set key/value.")


@etcd.command()
@click.argument("key", type=str, metavar="KEY")
@click.option(
    "-p",
    "--prefix",
    is_flag=True,
    default=False,
    help="Delete all keys prefixed with the given key.",
)
def delete(key, prefix):
    """
    Delete key(s) from ETCD.

    KEY: Name of ETCD key.
    """
    with Session() as session:
        try:
            data = session.EtcdConfig.delete(key, prefix)
        except Exception as e:
            print_error(e)
            sys.exit(ExitCode.FAILURE)
        if data.get("result", False) != "ok":
            print_fail("Unable to delete key/value.")
        else:
            print_pretty("Successfully deleted key/value.")



import textwrap
from collections.abc import Sequence
from decimal import Decimal
from typing import Any, Optional, Self
from uuid import UUID

from ai.backend.common.types import AutoScalingMetricComparator, AutoScalingMetricSource

from ...cli.types import Undefined, undefined
from ..output.fields import service_auto_scaling_rule_fields
from ..output.types import FieldSpec, RelayPaginatedResult
from ..pagination import execute_paginated_relay_query
from ..session import api_session
from ..types import set_if_set
from .base import BaseFunction, api_function

_default_fields: Sequence[FieldSpec] = (
    service_auto_scaling_rule_fields["id"],
    service_auto_scaling_rule_fields["metric_source"],
    service_auto_scaling_rule_fields["metric_name"],
    service_auto_scaling_rule_fields["comparator"],
    service_auto_scaling_rule_fields["threshold"],
    service_auto_scaling_rule_fields["endpoint"],
    service_auto_scaling_rule_fields["step_size"],
    service_auto_scaling_rule_fields["cooldown_seconds"],
    service_auto_scaling_rule_fields["min_replicas"],
    service_auto_scaling_rule_fields["max_replicas"],
    service_auto_scaling_rule_fields["created_at"],
    service_auto_scaling_rule_fields["last_triggered_at"],
)


class ServiceAutoScalingRule(BaseFunction):
    rule_id: UUID

    @api_function
    @classmethod
    async def paginated_list(
        cls,
        endpoint_id: UUID,
        *,
        fields: Sequence[FieldSpec] | None = None,
        page_offset: int = 0,
        page_size: int = 20,
        filter: Optional[str] = None,
        order: Optional[str] = None,
    ) -> RelayPaginatedResult[dict]:
        return await execute_paginated_relay_query(
            "endpoint_auto_scaling_rule_nodes",
            {
                "endpoint": (str(endpoint_id), "String!"),
                "filter": (filter, "String"),
                "order": (order, "String"),
            },
            fields or _default_fields,
            limit=page_size,
            offset=page_offset,
        )

    @api_function
    @classmethod
    async def create(
        cls,
        service: UUID,
        metric_source: AutoScalingMetricSource,
        metric_name: str,
        threshold: Decimal,
        comparator: AutoScalingMetricComparator,
        step_size: int,
        cooldown_seconds: int,
        *,
        min_replicas: Optional[int] = None,
        max_replicas: Optional[int] = None,
    ) -> Self:
        q = textwrap.dedent(
            """
            mutation(
                $endpoint: String!,
                $metric_source: AutoScalingMetricSource!,
                $metric_name: String!,
                $threshold: String!,
                $comparator: AutoScalingMetricComparator!,
                $step_size: Int!,
                $cooldown_seconds: Int!,
                $min_replicas: Int,
                $max_replicas: Int
            ) {
                create_endpoint_auto_scaling_rule_node(
                    endpoint: $endpoint,
                    props: {
                        metric_source: $metric_source,
                        metric_name: $metric_name,
                        threshold: $threshold,
                        comparator: $comparator,
                        step_size: $step_size,
                        cooldown_seconds: $cooldown_seconds,
                        min_replicas: $min_replicas,
                        max_replicas: $max_replicas
                    }
                ) {
                    rule {
                        row_id
                    }
                }
            }
            """
        )
        data = await api_session.get().Admin._query(
            q,
            {
                "endpoint": str(service),
                "metric_source": metric_source,
                "metric_name": metric_name,
                "threshold": threshold,
                "comparator": comparator,
                "step_size": step_size,
                "cooldown_seconds": cooldown_seconds,
                "min_replicas": min_replicas,
                "max_replicas": max_replicas,
            },
        )

        return cls(rule_id=UUID(data["create_endpoint_auto_scaling_rule_node"]["rule"]["row_id"]))

    def __init__(self, rule_id: UUID) -> None:
        super().__init__()
        self.rule_id = rule_id

    @api_function
    async def get(
        self,
        fields: Sequence[FieldSpec] | None = None,
    ) -> Sequence[dict]:
        query = textwrap.dedent(
            """\
            query($rule_id: String!) {
                endpoint_auto_scaling_rule_node(id: $rule_id) {$fields}
            }
        """
        )
        query = query.replace("$fields", " ".join(f.field_ref for f in (fields or _default_fields)))
        variables = {"rule_id": str(self.rule_id)}
        data = await api_session.get().Admin._query(query, variables)
        return data["endpoint_auto_scaling_rule_node"]

    @api_function
    async def update(
        self,
        *,
        metric_source: AutoScalingMetricSource | Undefined = undefined,
        metric_name: str | Undefined = undefined,
        threshold: Decimal | Undefined = undefined,
        comparator: AutoScalingMetricComparator | Undefined = undefined,
        step_size: int | Undefined = undefined,
        cooldown_seconds: int | Undefined = undefined,
        min_replicas: Optional[int] | Undefined = undefined,
        max_replicas: Optional[int] | Undefined = undefined,
    ) -> Self:
        q = textwrap.dedent(
            """
            mutation(
                $rule_id: String!,
                $input: ModifyEndpointAutoScalingRuleInput!,
            ) {
                modify_endpoint_auto_scaling_rule_node(
                    id: $rule_id,
                    props: $input
                ) {
                    ok
                    msg
                }
            }
            """
        )
        inputs: dict[str, Any] = {}
        set_if_set(inputs, "metric_source", metric_source)
        set_if_set(inputs, "metric_name", metric_name)
        set_if_set(inputs, "threshold", threshold)
        set_if_set(inputs, "comparator", comparator)
        set_if_set(inputs, "step_size", step_size)
        set_if_set(inputs, "cooldown_seconds", cooldown_seconds)
        set_if_set(inputs, "min_replicas", min_replicas)
        set_if_set(inputs, "max_replicas", max_replicas)
        data = await api_session.get().Admin._query(
            q,
            {"rule_id": str(self.rule_id), "input": inputs},
        )

        return data["modify_endpoint_auto_scaling_rule_node"]

    @api_function
    async def delete(self) -> None:
        q = textwrap.dedent(
            """
            mutation($rule_id: String!) {
                delete_endpoint_auto_scaling_rule_node(id: $rule_id) {
                    ok
                    msg
                }
            }
            """
        )

        variables = {
            "rule_id": str(self.rule_id),
        }
        data = await api_session.get().Admin._query(q, variables)
        return data["delete_endpoint_auto_scaling_rule_node"]
